@@ -27,11 +27,13 @@ static gchar *opt_path = ".";
 static gboolean opt_stream = FALSE;
 static gboolean opt_noprogress = FALSE;
 static gboolean opt_print_names = FALSE;
+static gboolean opt_choose_files = FALSE;
 
 static GOptionEntry entries[] = {
 	{ "path", '\0', 0, G_OPTION_ARG_FILENAME, &opt_path, "Local directory or file name, to save data to", "PATH" },
 	{ "no-progress", '\0', 0, G_OPTION_ARG_NONE, &opt_noprogress, "Disable progress bar", NULL },
 	{ "print-names", '\0', 0, G_OPTION_ARG_NONE, &opt_print_names, "Print names of downloaded files", NULL },
+	{ "choose-files", '\0', 0, G_OPTION_ARG_NONE, &opt_choose_files, "Choose which files to download when downloading folders (interactive)", NULL },
 	{ NULL }
 };
 
@@ -56,22 +58,37 @@ static void status_callback(struct mega_status_data *data, gpointer userdata)
 
 // download operation
 
-static gboolean dl_sync_file(struct mega_node *node, GFile *file, const gchar *remote_path)
+static gboolean dl_sync_file(struct mega_node *node, GFile *file)
 {
 	gc_error_free GError *local_err = NULL;
+	gc_object_unref GFile *parent = g_file_get_parent(file);
 	gc_free gchar *local_path = g_file_get_path(file);
+	gc_free gchar *parent_path = g_file_get_path(parent);
 
 	if (g_file_query_exists(file, NULL)) {
 		g_printerr("ERROR: File already exists at %s\n", local_path);
 		return FALSE;
 	}
 
+	if (!g_file_query_exists(parent, NULL)) {
+		if (!g_file_make_directory_with_parents(parent, NULL, &local_err)) {
+			g_printerr("ERROR: Can't create local directory %s: %s\n", parent_path, local_err->message);
+			return FALSE;
+		}
+	} else {
+		if (g_file_query_file_type(parent, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL) != G_FILE_TYPE_DIRECTORY) {
+			g_printerr("ERROR: Can't create local directory %s: a file exists there!\n", parent_path);
+			return FALSE;
+		}
+	}
+
 	if (!opt_noprogress)
 		g_print("F %s\n", local_path);
 
-	if (!mega_session_get_compat(s, local_path, remote_path, &local_err)) {
+	if (!mega_session_get(s, file, node, &local_err)) {
 		if (!opt_noprogress && tool_is_stdout_tty())
 			g_print("\r" ESC_CLREOL);
+		gc_free gchar* remote_path = mega_node_get_path_dup(node);
 		g_printerr("ERROR: Download failed for %s: %s\n", remote_path, local_err->message);
 		return FALSE;
 	}
@@ -85,39 +102,23 @@ static gboolean dl_sync_file(struct mega_node *node, GFile *file, const gchar *r
 	return TRUE;
 }
 
-static gboolean dl_sync_dir(struct mega_node *node, GFile *file, const gchar *remote_path)
+static gboolean dl_sync_dir(struct mega_node *node, GFile *file)
 {
 	gc_error_free GError *local_err = NULL;
 	gc_free gchar *local_path = g_file_get_path(file);
 
-	if (!g_file_query_exists(file, NULL)) {
-		if (!opt_noprogress)
-			g_print("D %s\n", local_path);
-
-		if (!g_file_make_directory(file, NULL, &local_err)) {
-			g_printerr("ERROR: Can't create local directory %s: %s\n", local_path, local_err->message);
-			return FALSE;
-		}
-	} else {
-		if (g_file_query_file_type(file, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL) != G_FILE_TYPE_DIRECTORY) {
-			g_printerr("ERROR: Can't create local directory %s: file exists\n", local_path);
-			return FALSE;
-		}
-	}
-
 	// sync children
 	GSList *children = mega_session_get_node_chilren(s, node), *i;
 	gboolean status = TRUE;
 	for (i = children; i; i = i->next) {
 		struct mega_node *child = i->data;
-		gc_free gchar *child_remote_path = g_strconcat(remote_path, "/", child->name, NULL);
 		gc_object_unref GFile *child_file = g_file_get_child(file, child->name);
 
-		if (child->type == 0) {
-			if (!dl_sync_file(child, child_file, child_remote_path))
+		if (child->type == MEGA_NODE_FILE) {
+			if (!dl_sync_file(child, child_file))
 				status = FALSE;
 		} else {
-			if (!dl_sync_dir(child, child_file, child_remote_path))
+			if (!dl_sync_dir(child, child_file))
 				status = FALSE;
 		}
 	}
@@ -126,6 +127,195 @@ static gboolean dl_sync_dir(struct mega_node *node, GFile *file, const gchar *re
 	return status;
 }
 
+static gint* parse_number_list(const gchar* input, gint* count)
+{
+	*count = 0;
+
+	gc_regex_unref GRegex* re = g_regex_new("\\s+", 0, 0, NULL);
+	if (re == NULL)
+		return NULL;
+
+	gchar** tokens = g_regex_split(re, input, 0);
+	guint i = 0, n_tokens = g_strv_length(tokens);
+	gint* nums = g_new0(gint, n_tokens);
+
+	*count = 0;
+        while (i < n_tokens) {
+		if (g_regex_match_simple("^\\d{1,7}$", tokens[i], 0, 0)) {
+			nums[*count] = atoi(tokens[i]);
+			*count += 1;
+		} else {
+			g_printerr("WARNING: Skipping non-numeric value '%s'\n", tokens[i]);
+		}
+
+		i++;
+	}
+
+	g_strfreev(tokens);
+	return nums;
+}
+
+static GSList *prompt_and_filter_nodes(GSList *nodes)
+{
+	GSList *chosen_nodes = NULL;
+
+        gc_free gchar* input = tool_prompt_input();
+	if (input == NULL)
+		return g_slist_copy(nodes);
+
+	if (g_regex_match_simple("all", input, G_REGEX_CASELESS, 0))
+		return g_slist_copy(nodes);
+
+	gint n_nums;
+	gc_free gint* nums = parse_number_list(input, &n_nums);
+
+	gint max_index = g_slist_length(nodes);
+
+	for (int i = 0; i < n_nums; i++) {
+		if (nums[i] <= max_index && nums[i] >= 1) {
+			struct mega_node* node = g_slist_nth_data(nodes, nums[i] - 1);
+
+			if (!g_slist_find(chosen_nodes, node))
+				chosen_nodes = g_slist_prepend(chosen_nodes, node);
+			else
+				g_printerr("WARNING: Index %d given multiple times\n", nums[i]);
+		} else
+			g_printerr("WARNING: Index %d out of range\n", nums[i]);
+	}
+
+	return chosen_nodes;
+}
+
+static gint compare_node(struct mega_node *a, struct mega_node *b)
+{
+	gchar path1[4096];
+	gchar path2[4096];
+
+	if (mega_node_get_path(a, path1, 4096) && mega_node_get_path(b, path2, 4096)) {
+		if (mega_node_is_container(a)) {
+			int pos = strlen(path1);
+			if(pos < 4095) {
+				path1[pos] = '/';
+				path1[pos+1] = '\0';
+			}
+		}
+		if (mega_node_is_container(b)) {
+			int pos = strlen(path2);
+			if(pos < 4095) {
+				path2[pos] = '/';
+				path2[pos+1] = '\0';
+			}
+		}
+		return strcmp(path1, path2);
+	}
+	return 0;
+}
+
+GSList* prune_children(GSList* nodes)
+{
+	GSList* pruned = NULL, *it, *it2;
+
+	// not the most optimal, but the working set is small
+	for (it = nodes; it; it = it->next) {
+		struct mega_node *node = it->data;
+
+		// check if there's ancestor of node in the list
+		for (it2 = nodes; it2; it2 = it2->next) {
+			struct mega_node *node2 = it2->data;
+
+			if (mega_node_has_ancestor(node, node2)) {
+				gchar path[4096];
+				if (mega_node_get_path(node, path, sizeof path))
+					g_printerr("WARNING: skipping already included path %s\n", path);
+
+				goto prune_node;
+			}
+		}
+
+		pruned = g_slist_prepend(pruned, node);
+
+prune_node:;
+	}
+
+	g_slist_free(nodes);
+
+	return g_slist_reverse(pruned);
+}
+
+GSList* pick_nodes(void)
+{
+	GSList *nodes = mega_session_ls(s, "/", TRUE), *it, *chosen_nodes;
+	int position = 2;
+
+	nodes = g_slist_sort(nodes, (GCompareFunc)compare_node);
+
+	struct mega_node *parent = nodes->data;
+	int indent = 0;
+
+	g_print("1. %s/\n", ((struct mega_node*)nodes->data)->name);
+
+	for (it = nodes->next; it; it = it->next) {
+		struct mega_node *node = it->data;
+		while(parent != node->parent && indent != 0) {
+			indent--;
+			parent = parent->parent;
+		}
+
+		g_print("%*s" ESC_GRAY "|--" ESC_NORMAL "%d. %s", indent*3, "", position, node->name);
+		if (mega_node_is_container(node)) {
+			g_print("/");
+			indent++;
+			parent = node;
+		}
+		g_print("\n");
+
+		position++;
+	}
+
+	g_print("Enter numbers of files or folders to download separated by spaces (or type 'all' to download everything):\n> ");
+
+	chosen_nodes = prompt_and_filter_nodes(nodes);
+
+	g_slist_free(nodes);
+
+	// now, user may have chosen both directories and files contained in
+	// them, prune the list
+
+	chosen_nodes = g_slist_sort(chosen_nodes, (GCompareFunc)compare_node);
+
+	return prune_children(chosen_nodes);
+}
+
+static gboolean dl_sync_dir_choose(GFile *local_dir)
+{
+	gc_error_free GError *local_err = NULL;
+	GSList* chosen_nodes = pick_nodes(), *it;
+	gboolean status = TRUE;
+
+	if (chosen_nodes == NULL)
+		g_printerr("WARNING: Nothing was selected\n");
+
+	for (it = chosen_nodes; it; it = it->next) {
+		struct mega_node *node = it->data;
+		gchar remote_path[4096];
+		if (!mega_node_get_path(node, remote_path, sizeof remote_path))
+			continue;
+
+		gc_object_unref GFile *file = g_file_get_child(local_dir, remote_path + 1);
+
+		if (node->type == MEGA_NODE_FILE) {
+			if (!dl_sync_file(node, file))
+				status = FALSE;
+		} else {
+			if (!dl_sync_dir(node, file))
+				status = FALSE;
+		}
+	}
+
+	g_slist_free(chosen_nodes);
+	return status;
+}
+
 int main(int ac, char *av[])
 {
 	gc_error_free GError *local_err = NULL;
@@ -133,7 +323,8 @@ int main(int ac, char *av[])
 	gint i;
 	int status = 0;
 
-	tool_init(&ac, &av, "- download exported files from mega.nz", entries, TOOL_INIT_AUTH_OPTIONAL | TOOL_INIT_DOWNLOAD_OPTS);
+	tool_init(&ac, &av, "- download exported files from mega.nz", entries,
+		  TOOL_INIT_AUTH_OPTIONAL | TOOL_INIT_DOWNLOAD_OPTS);
 
 	if (!strcmp(opt_path, "-")) {
 		opt_noprogress = opt_stream = TRUE;
@@ -237,9 +428,13 @@ int main(int ac, char *av[])
 					gc_object_unref GFile *local_dir = g_file_new_for_path(opt_path);
 					if (g_file_query_file_type(local_dir, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
 								   NULL) == G_FILE_TYPE_DIRECTORY) {
-						gc_free gchar *node_path = mega_node_get_path_dup(root_node);
-						if (!dl_sync_dir(root_node, local_dir, node_path))
-							status = 1;
+						if (opt_choose_files) {
+							if (!dl_sync_dir_choose(local_dir))
+								status = 1;
+						} else {
+							if (!dl_sync_dir(root_node, local_dir))
+								status = 1;
+						}
 					} else {
 						g_printerr("ERROR: %s must be a directory\n", opt_path);
 						status = 1;
